import tensorflow as tf


def _int64_feature(values):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=values))


def _float_feature(values):
    return tf.train.Feature(float_list=tf.train.FloatList(value=values))


def _bytes_feature(values):
    if isinstance(values[0], str):
        values = values.astype(bytes)
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=values))


class TFExample(object):
    def __init__(self, input_features):
        self.convert_funcs = {}
        self.feature_specs = {}
        self.input_features = input_features.copy()
        for k, fea in input_features.items():
            if fea.dtype in [tf.int8, tf.int16, tf.int32, tf.int32, tf.int64]:
                self.convert_funcs[fea.name] = _int64_feature
                dtype = tf.int64
            elif fea.dtype in [tf.float32, tf.float64]:
                self.convert_funcs[fea.name] = _float_feature
                dtype = tf.float32
            elif fea.dtype in [tf.string]:
                self.convert_funcs[fea.name] = _bytes_feature
                dtype = tf.string
            self.feature_specs[fea.name] = tf.io.FixedLenFeature(fea.shape, dtype)

    def serialize(self, example):
        for k, v in example.items():
            example[k] = self.convert_funcs[k](v.flatten())
        example_proto = tf.train.Example(features=tf.train.Features(feature=example))
        return example_proto.SerializeToString()

    def deserialize(self, example_proto):
        return tf.io.parse_single_example(example_proto, self.feature_specs)


def gather_seqs(seqs, gather_inds):
    dtype = gather_inds.dtype
    n = len(gather_inds.get_shape())
    shape = tf.shape(gather_inds)
    r_shape = [shape[0]]
    inds=tf.range(shape[0]); tile=[1]
    inds = tf.expand_dims(inds, -1)
    for i in range(1, n-1):
        inds=tf.expand_dims(inds, i)
        inds = tf.tile(inds, tile+[shape[i]] +[1])
        r_shape.append(shape[i])
        r_inds = tf.range(shape[i])
        r_inds = tf.expand_dims(r_inds, -1)
        for j in tile[::-1]:
            r_inds = tf.expand_dims(r_inds, 0)
        r_inds = r_inds * tf.ones(r_shape+[1], tf.int32)
        inds = tf.concat([inds, r_inds], -1)
        tile += [1]
    inds = tf.tile(tf.expand_dims(inds, -2), tile + [shape[-1]] + [1])
    gather_inds = tf.expand_dims(gather_inds, -1)
    gather_inds = tf.concat([inds, gather_inds], -1)
    rst = tf.gather_nd(seqs, gather_inds)
    return rst

def flat_gradients(grads_or_idx_slices: tf.Tensor) -> tf.Tensor:
    '''Convert gradients if it's tf.IndexedSlices.
    When computing gradients for operation concerning `tf.gather`, the type of gradients
    '''
    if type(grads_or_idx_slices) == tf.IndexedSlices:
        return tf.scatter_nd(
            tf.expand_dims(grads_or_idx_slices.indices, 1),
            grads_or_idx_slices.values,
            grads_or_idx_slices.dense_shape
        )
    return grads_or_idx_slices



import tensorflow as tf
from tensorflow.keras.layers import Conv2D, Conv2DTranspose

from mautil.tf import NN

class Block(NN):
    def __init__(self, name='Block', filters=64, kernel_size=3, strides=(1,1), activation='relu', padding='same', batch_norm=True,
                 transpose=False, **kwargs):
        super(ResidualBlock, self).__init__(name=name, **kwargs)
        self.net = tf.keras.Sequential()
        if batch_norm:
            self.net.add(tf.keras.layers.BatchNormalization())
        self.net.add(tf.keras.layers.Activation(activation))
        if transpose:
            self.net.add(Conv2DTranspose(filters, kernel_size, strides=strides, activation=activation, padding=padding))
        else:
            self.net.add(Conv2D(filters, kernel_size, strides=strides, activation=activation, padding=padding))

    def call(self, inputs, training=None, mask=None):
        return self.net(inputs, training=training, mask=mask)


class ResidualBlock(NN):
    def __init__(self, name='ResidualBlock', filters=64, kernel_size=3, activation='relu', padding='same', reduction_factor=4,
                 downsample=True, upsample=False, strides=(2,2), batch_norm=True, **kwargs):
        """
        conv0 for skip connection
        :param name:
        :param filters:
        :param kernel_size:
        :param activation:
        :param padding:
        :param reduction_factor:
        :param downsample:
        :param upsample:
        :param strides:
        :param batch_norm:
        :param kwargs:
        """
        super(ResidualBlock, self).__init__(name=name, **kwargs)

        self._downsample = downsample
        self._upsample = upsample

        self.conv1 = Block(filters//reduction_factor,(1,1), (1,1), activation=activation, batch_norm=batch_norm, padding=padding)
        if upsample:
            self.conv2 = Block(filters // reduction_factor, (kernel_size, kernel_size), strides=strides, activation=activation, batch_norm=batch_norm, padding=padding, transpose=True)
            self.conv0 =Block(filters, (kernel_size, kernel_size), strides=strides, activation=None, batch_norm=False, padding=padding, transpose=True)
        else:
            self.conv2 = Block(filters//reduction_factor,(kernel_size, kernel_size), (1,1), activation=activation, batch_norm=batch_norm, padding=padding)
        self.conv3 = Block(filters//reduction_factor,(kernel_size, kernel_size), (1,1), activation=activation, batch_norm=batch_norm, padding=padding)
        if downsample:
            self.conv4 = Block(filters, (1,1), strides=strides, activation=activation, batch_norm=batch_norm)
            self.conv0 = Block(filters, (kernel_size, kernel_size), strides=strides, activation=None, batch_norm=False)
        else:
            self.conv4 = Block(filters, (1,1), strides=(1,1), activation=activation, batch_norm=batch_norm)

    def call(self, inputs, training=None, mask=None):
        out1 = self.conv1(inputs, training=training, mask=mask)
        out2 = self.conv1(out1, training=training, mask=mask)
        out3 = self.conv1(out2, training=training, mask=mask)
        out4 = self.conv1(out3, training=training, mask=mask)
        if self._downsample:
            iniputs = self.conv0(inputs, training, mask=mask)
        out = inputs + out4
        return out









import os, sys, logging, json, time, re
import pickle
import numpy as np
import psutil
import math
import inspect
import argparse
from collections import OrderedDict
from contextlib import contextmanager
from collections import abc


logger = logging.getLogger(__name__)


class NestedObject(argparse.Namespace):
    def __init__(self, d):
        for a, b in d.items():
            if isinstance(b, (list, tuple)):
               setattr(self, a, [NestedObject(x) if isinstance(x, dict) else x for x in b])
            else:
               setattr(self, a, NestedObject(b) if isinstance(b, dict) else b)


def update_dict(d, u):
    for k, v in u.items():
        if isinstance(v, abc.Mapping):
            d[k] = update_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d

def get_default_args(func):
    signature = inspect.signature(func)
    return { k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}


def fill_kwargs(func, args, kwargs):
    default_args = get_default_args(func)
    default_args.update(kwargs)
    arg_names = [k for k, v in inspect.signature(func).parameters.items() if k!='self']
    for k, v in zip(arg_names, args):
        default_args[k] = v
    return default_args


def dynamic_import(kls):
    parts = kls.split('.')
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)
    return m


def run_cmd(cmd, echo=True, asserting=True):
    if echo:
        logger.info('cmd is %s', cmd)
    rst = os.system(cmd)
    if asserting:
        assert rst == 0

def df2dict(df):
    data = {}
    for k, v in df.to_dict('list').items():
        data[k] = np.array(v)
    return data


def onehot(labels, dim):
    """
    :param labels: one dimension array of int
    :param dim: one hot dimension, i.e. total num of class
    :return:
    """
    num = len(labels)
    onehot = np.zeros((num, dim), np.int32)
    onehot[np.arange(num), labels] = 1
    return onehot


def get_default_num_job():
    from multiprocessing import cpu_count
    return max(1, cpu_count()-1)


def load_dump(fpath):
    with open(fpath, 'rb') as f:
        data = pickle.load(f)
    return data


def dump(data, fpath, protocol=2):
    fdir = os.path.dirname(fpath)
    if not os.path.exists(fdir):
        os.makedirs(fdir)
    with open(fpath, 'wb') as f:
        pickle.dump(data, f)


def load_json(fpath):
    with open(fpath) as f:
        dictionary = json.load(f)
    return dictionary


def load_json_lines(fpath):
    data = []
    with open(fpath) as f:
        for l in f:
            data.append(json.loads(l))
    return data


def dump_json(dictionary, fpath, ensure_ascii=False):
    with open(fpath, 'w') as f:
        json.dump(dictionary, f, ensure_ascii=ensure_ascii)


def dump_json_lines(dicts, fpath, ensure_ascii=False):
    with open(fpath, 'w', encoding='utf8') as f:
        for d in dicts:
            json.dump(d, f, ensure_ascii=ensure_ascii)
            f.write(os.linesep)


def pad_sequence(seqs, pad=0, seqs_len=None, dtype=np.int32, max_seq_len=None):
    seqs = np.array(seqs)
    shape = seqs.shape
    if seqs.dtype == object:
        flatten_seqs = seqs.flatten()
        seqs_len = np.reshape([len(x) for x in flatten_seqs], shape)
        seqs = np.concatenate(flatten_seqs)
        if max_seq_len is None:
            max_seq_len = np.max(seqs_len)
        mask = np.arange(max_seq_len) < np.expand_dims(seqs_len, -1)
        padded_seqs = np.ones(mask.shape, dtype) * pad
        padded_seqs[mask] = seqs
    else:
        padded_seqs = seqs
        shape = padded_seqs.shape
        pads = []
        if max_seq_len is not None:
            assert shape[-1]<=max_seq_len
            if shape[-1] < max_seq_len:
                for dim in shape[:-1]:
                    pads.append((0, 0))
                pads.append((0, max_seq_len-shape[-1]))
                padded_seqs = np.pad(padded_seqs, tuple(pads), 'constant', constant_values=pad)
        seqs_len = np.ones(shape[:-1], dtype)*shape[-1]
    return padded_seqs, seqs_len


def timestamp():
    return time.strftime('%Y%m%d%H%M%S')


@contextmanager
def timer(name):
    t0 = time.time()
    #print('{} start'.format(name))
    logger.info('%s start', name)
    yield
    #print('{} done in {} seconds'.format(name, time.time() - t0))
    logger.info('%s done in %s seconds', name, time.time()-t0)


@contextmanager
def trace(name, sleep=1):
    t0 = time.time()
    p = psutil.Process(os.getpid())
    m0 = p.memory_info()[0] / 2. ** 20
    #print('{} start'.format(name))
    logger.info('%s start', name)
    yield
    time.sleep(sleep)
    m1 = p.memory_info()[0] / 2. ** 20

    #print('{} done in {} seconds'.format(name, time.time() - t0))
    logger.info('%s done in %s seconds, memory percent:%s, delta:%sM', name, time.time()-t0, psutil.virtual_memory().percent, m1-m0)


def merge_args(args, cfg={}):
    for k, v in args.__dict__.items():
        if v is not None:
            cfg[k] = v
    return cfg


def parse_model_name(model_names):
    names2cls = OrderedDict()
    for name in re.split('[, ]+', model_names):
        names2cls[name] = name.split('_')[0]
    return names2cls


def create_model(name, args, cls, suffix='', cfg={}, **kwargs):
    name = name + suffix
    cfg = cfg.copy()
    cfg = merge_args(args, cfg)
    model = cls(name=name, cfg=cfg, **kwargs)
    return model


def set_logger(level=logging.INFO):
    logger = logging.getLogger()
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s -   %(message)s')
    handler.setFormatter(formatter)
    logger.handlers = [handler]
    logger.setLevel(level)

def unpack_args(func):
    from functools import wraps
    @wraps(func)
    def wrapper(args):
        if isinstance(args, dict):
            return func(**args)
        else:
            return func(*args)
    return wrapper


def update_deepspeed_cfg(cfg):
    if not cfg.use_deepspeed:
        del cfg.deepspeed
        return
    config = cfg.deepspeed
    if not cfg.use_fp16:
        logger.warning('automatically enable fp16 for deepspeed')
    cfg.use_fp16 = False
    config['fp16']['enabled'] = True

    config['train_micro_batch_size_per_gpu'] = cfg.batch_size
    config['gradient_accumulation_steps'] = cfg.accumulated_batch_size
    if cfg.gradient_clip is not None:
        config['gradient_clipping'] = cfg.gradient_clip
    #config['optimizer']['type'] = cfg.opt
    config['optimizer']['params'] = cfg.opt_paras
    config['optimizer']['params']['lr'] = cfg.lr

    scheduler = "WarmupLR"
    params = {
        "warmup_min_lr": 0,
        "warmup_max_lr": cfg.lr,
        "warmup_num_steps": cfg.n_lr_warmup_step,
    }
    config["scheduler"] = {
        "type": scheduler,
        "params": params,
    }


def init_deepspeed(config, model):
    class Object(object):
        pass
    args = Object()
    args.local_rank = -1
    import deepspeed
    model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    model, optimizer, _, lr_scheduler = deepspeed.initialize(
        args=args,  # expects an obj
        model=model,
        model_parameters=model_parameters,
        config_params=config,
    )
    print(111, optimizer.optimizer.state_dict().keys())
    logger.info('deepspeed model initialized')

    return model, optimizer, lr_scheduler


## arg parser
class ArgParser(argparse.ArgumentParser):
    def __init__(self, **kwargs):
        super(ArgParser, self).__init__(prog=os.path.basename(sys.argv[0]),
                                         formatter_class=argparse.RawDescriptionHelpFormatter, description=__doc__, **kwargs)

        self.add_argument("-m", "--method_name", help="run method name")
        self.add_argument("-num", type=int, default=int(1e16), help="num of examples")
        self.add_argument("-d", "--debug", action="store_true", help="debug")
        self.add_argument("-data_dir", default='../data', help="data dir")
        self.add_argument("-output_dir", default='../data', help="output dir")
        self.add_argument("-model_dir")
        self.add_argument("-save", action="store_true", help="save")


class TrainArgParser(ArgParser):
    def __init__(self, **kwargs):
        super(TrainArgParser, self).__init__(**kwargs)

        self.add_argument("-seed", type=int, default=9527)
        self.add_argument("-show_cfg", action="store_true", help="show model cfg")
        self.add_argument("-ms", "--model_names", help="model names")
        self.add_argument("-scoring", "--scoring", action="store_true", help="run score after train")
        self.add_argument("-ov", "--only_validate", action="store_true", help="only do validation")
        self.add_argument("-nt", "--no_train", action="store_true", help="do not do train")
        self.add_argument("-np", "--no_predicting", action="store_true", help="do not do predicting")
        self.add_argument("-ntest", "--no_test", action="store_true", help="do not pred test")
        self.add_argument("-predicting", "--predicting", action="store_true", help="run predict after train")
        self.add_argument("-brcfg", "--batch_reader_cfg", type=json.loads, help="data reader cfg")
        self.add_argument("-tf_loglevel", "--tf_loglevel", type=int, default=1, help="tf log level.0:DEBUG, 1:INFO, 2:WARN, 3: ERROR")
        self.add_argument("-use_tf_estimator", "--use_tf_estimator", action="store_true", help="use tensorflow estimator api")
        self.add_argument("-regen_tfrecord", "--regen_tfrecord", action="store_true", help="regeneration of tf records")
        self.add_argument("-fake_colab", "--fake_colab", action="store_true", help="used for debug at local")
        self.add_argument("-warm_start_path", "--warm_start_path", help="tpu warm start")
        self.add_argument("-tfrecords_repeat", type=int, help="when generate tf records, repeat the data")
        self.add_argument("-num_core_per_host", type=int,  help="num_core_per_host")
        self.add_argument("-save_opt", action="store_true", help="save optimizer")
        self.add_argument("-save_half", action="store_true", help="save half precision(fp16)")
        self.add_argument("-save_best", action="store_true", help="save best epoch")
        self.add_argument("-save_log", action="store_true", help="save log to file")
        self.add_argument("-save_processed", "--save_processed", action="store_true", help="save processed data")
        self.add_argument("-save_epoch", "--save_epoch", type=int, help="save per epochs")

        # dataset
        self.add_argument("-shuffle_buffer_size", type=int, default=128, help="dataset shuffle buffer size")
        self.add_argument("-n_dl_worker", type=int, help="data loader works")

        # gpu
        self.add_argument("-device", nargs='+', type=int, help="specify gpus")
        self.add_argument("-use_mg", "--use_multiple_gpu", action="store_true", help="enable multiple gpu if available")

        # tpu
        self.add_argument("-use_tpu", "--use_tpu", action="store_true", help="use tpu")
        self.add_argument("-tpu_name", "--tpu_name",  help="tpu name")
        self.add_argument("-gsurl", "--gsurl", help="google cloud storage url for save tfrecords and model, used for tpu")
        self.add_argument("-tpu_zone", "--tpu_zone",  help="tpu zone")
        self.add_argument("-tpu_loop_iterations", type=int,  help="tpu loop iterations")
        self.add_argument("-gcp_project", "--gcp_project",  help="gcp project")

        # xla
        self.add_argument("-xla_procs", type=int,  help="num of process for xla_multiprocessing")
        self.add_argument("-xla_spawn_method", help="spawn method for xla_multiprocessing")
        self.add_argument("-xla_share_model", action="store_true", help="share model for multiprocessing to save memory")



        # training
        self.add_argument("-dp", "--dropout", type=float, default=0.0)
        self.add_argument("-save_pred", action="store_true", help="save pred")
        self.add_argument("-use_fp16", action="store_true", help="mixed precision fp16 training")
        self.add_argument("-use_deepspeed", action="store_true")
        self.add_argument("-recompute", action="store_true", help="recompute grads")
        self.add_argument("-run_eagerly", action="store_true", help="tenesorflow eager mode")
        self.add_argument("-bs", "--batch_size", type=int, default=2, help="batch size")
        self.add_argument("-vbs", "--val_batch_size", type=int, default=2, help="validate batch size")
        self.add_argument("-abs", "--accumulated_batch_size", type=int, help="accumulated batch training")
        self.add_argument("-n_save_step", type=int, help="save per steps")
        self.add_argument("-n_save_epoch", type=int, help="save per epochs")
        self.add_argument("-n_init_epoch", type=int, help="ignore such as validation stuff for first n epochs")
        self.add_argument("-n_es_epoch", type=int, help="early stop after n epochs without improve")
        self.add_argument("-restore_epoch", type=int, help="restore epoch training from")
        self.add_argument("-n_epoch_step", type=int, help="num of steps per train epoch")
        self.add_argument("-nv", "--no_validate", action="store_true", help="do not do validation")
        self.add_argument("-validation_steps", type=int)
        self.add_argument("-validation_frea", type=int)
        self.add_argument("-initial_epoch", type=int, default=0)
        self.add_argument("-patience", type=int, help='Number of epochs with no improvement after which training will be stopped')
        self.add_argument("-es", "--es_min_delta", type=float, help="Minimum change in the monitored quantity to qualify as an improvement, i.e. an absolute change of less than min_delta, will count as no improvement.")
        self.add_argument("-ema", type=float, help="if >0 use exponential moving average of the weights")
        self.add_argument("-save_ema", action="store_true", help="save the ema model seperately, usefully for resume training")
        self.add_argument("-swa_start_epoch", type=int, help="if >0 use stochastic weight average. pt only")
        self.add_argument("-swa_bs", "--swa_batch_size", type=int, help="swa batch size for bn update. pt only")
        self.add_argument("-swa_disable_bn", action="store_true", help="not update batch norm statistics for swa training. pt only")
        ## deepspeed
        self.add_argument("-local_rank", "--local_rank", type=int, default=-1)

        # save and  restore
        self.add_argument("-n_keep_ckpt", type=int, help="num of saved ckpts to keep")
        self.add_argument("-restore", action="store_true", help="restore")
        self.add_argument("-restore_strict", action="store_true", help="stric to exactly match state dict")
        self.add_argument("-not_restore_cfg", action="store_true", default=False, help="not restore cfg")

        # self distill
        self.add_argument("-sd_weight", type=float, help="self distill loss weight")
        self.add_argument("-sd_warmup_step", type=int, help="self distill warmup step")

        ## lr
        self.add_argument("-lr", "--lr", type=float, help="learning rate")
        self.add_argument("-n_lr_warmup_step", type=int, help="lr warmup steps")
        self.add_argument("-n_lr_decay_step", type=int, help="lr decay steps")
        self.add_argument("-lr_decay_rate", type=float, help="lr decay ratio")
        self.add_argument("-lr_scheduler", help='lr scheduler')
        self.add_argument("-lr_scheduler_paras", type=json.loads, help='lr scheduler parameters')
        self.add_argument("-lr_scheduler_ca_paras", type=json.loads, help='lr scheduler parameters for ca')
        self.add_argument("-lr_scheduler_ld_paras", type=json.loads, help='lr scheduler parameters for ld')


        self.add_argument("-save_summary_step", "--save_summary_step", type=int, help="save per steps")
        self.add_argument("-save_keep", "--save_keep", type=int, help="keep number of saved copy")
        self.add_argument("-weight_decay", type=float, help="weight decay")
        self.add_argument("-opt", help="optimizer:adam, radam, adamW, sgd")
        self.add_argument("-opt_paras", type=json.loads, help='parameters for optimizer')
        self.add_argument("-kernel_initializer", help="kernel initializer")
        self.add_argument("-gradient_clip", type=float, help="global norm gradient clip")
        self.add_argument("-warmup_steps", type=int, help="lr warm up steps")
        self.add_argument("-val_step", "--val_step", type=int, help="do validate per steps")
        self.add_argument("-val_batchs", "--val_batchs", type=int, help="only run  limited batchs when do validate per steps")
        self.add_argument("-ckpt_hours", "--keep_checkpoint_every_n_hours", type=int, help="")
        self.add_argument("-val_test", "--val_test", action="store_true", help="run validation on test split")
        self.add_argument("-fsuffix", default='', help="prefix name")
        self.add_argument("-fprefix", default='', help="prefix of generated name")
        self.add_argument("-l2", "--l2", type=float, help="l2 regularize")
        self.add_argument("-kfid", "--kfid", help="k fold ID")
        self.add_argument("-kn", "--kn", type=int, help="k fold num")
        self.add_argument("-epochs", "--epochs", type=int, help="training epochs")
        self.add_argument("-tp", "--train_pct", type=float, help="data percent used for train")
        self.add_argument("-verbose", "--verbose", type=int, help="verbose per num of batch")




# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc

import tensorflow as tf
from tensorflow.python.eager import context
from tensorflow.python.framework import ops
from tensorflow.python.ops import math_ops
from typing import Union


class AveragedOptimizerWrapper(tf.keras.optimizers.Optimizer, metaclass=abc.ABCMeta):
    def __init__(
            self,
            optimizer: Union[tf.keras.optimizers.Optimizer, str],
            sequential_update: bool = True,
            name: str = "AverageOptimizer",
            **kwargs
    ):
        super().__init__(name, **kwargs)

        if isinstance(optimizer, str):
            optimizer = tf.keras.optimizers.get(optimizer)

        if not isinstance(optimizer, tf.keras.optimizers.Optimizer):
            raise TypeError(
                "optimizer is not an object of tf.keras.optimizers.Optimizer"
            )

        if not isinstance(sequential_update, bool):
            raise TypeError("sequential_update must be of bool type")

        self._optimizer = optimizer
        self._sequential_update = sequential_update

    def _create_slots(self, var_list):
        self._optimizer._create_slots(var_list=var_list)
        for var in var_list:
            self.add_slot(var, "average")

    def _create_hypers(self):
        self._optimizer._create_hypers()

    def _prepare(self, var_list):
        return self._optimizer._prepare(var_list=var_list)

    def apply_gradients(self, grads_and_vars, name=None):
        self._optimizer._iterations = self.iterations
        return super().apply_gradients(grads_and_vars, name)

    @abc.abstractmethod
    def average_op(self, var, average_var):
        raise NotImplementedError

    def _apply_average_op(self, train_op, var):
        average_var = self.get_slot(var, "average")
        if self._sequential_update:
            with tf.control_dependencies([train_op]):
                avg_op = self.average_op(var, average_var)
        else:
            avg_op = self.average_op(var, average_var)

        return avg_op

    def _resource_apply_dense(self, grad, var):
        train_op = self._optimizer._resource_apply_dense(grad, var)
        average_op = self._apply_average_op(train_op, var)
        return tf.group(train_op, average_op)

    def _resource_apply_sparse(self, grad, var, indices):
        train_op = self._optimizer._resource_apply_sparse(grad, var, indices)
        average_op = self._apply_average_op(train_op, var)
        return tf.group(train_op, average_op)

    def _resource_apply_sparse_duplicate_indices(self, grad, var, indices):
        train_op = self._optimizer._resource_apply_sparse_duplicate_indices(
            grad, var, indices
        )
        average_op = self._apply_average_op(train_op, var)
        return tf.group(train_op, average_op)

    def assign_average_vars(self, var_list):
        """Assign variables in var_list with their respective averages.
        Args:
            var_list: List of model variables to be assigned to their average.
        Returns:
            assign_op: The op corresponding to the assignment operation of
            variables to their average.
        Example:
        ```python
        model = tf.Sequential([...])
        opt = tfa.optimizers.SWA(
                tf.keras.optimizers.SGD(lr=2.0), 100, 10)
        model.compile(opt, ...)
        model.fit(x, y, ...)
        # Update the weights to their mean before saving
        opt.assign_average_vars(model.variables)
        model.save('model.h5')
        ```
        """
        assign_op = tf.group(
            [
                var.assign(self.get_slot(var, "average"))
                for var in var_list
                if var.trainable
            ]
        )
        return assign_op

    def get_config(self):
        config = {
            "optimizer": tf.keras.optimizers.serialize(self._optimizer),
            "sequential_update": self._sequential_update,
        }
        base_config = super().get_config()
        return {**base_config, **config}

    @classmethod
    def from_config(cls, config, custom_objects=None):
        optimizer = tf.keras.optimizers.deserialize(
            config.pop("optimizer"), custom_objects=custom_objects,
        )
        return cls(optimizer, **config)

    @property
    def weights(self):
        return self._weights + self._optimizer.weights

    @property
    def lr(self):
        return self._optimizer._get_hyper("learning_rate")

    @lr.setter
    def lr(self, lr):
        self._optimizer._set_hyper("learning_rate", lr)  #

    @property
    def learning_rate(self):
        return self._optimizer._get_hyper("learning_rate")

    @learning_rate.setter
    def learning_rate(self, learning_rate):
        self._optimizer._set_hyper("learning_rate", learning_rate)


# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf
from tensorflow.python.training.moving_averages import assign_moving_average

from typing import Optional, Union


#@tf.keras.utils.register_keras_serializable(package='Addons')
class MovingAverage(AveragedOptimizerWrapper):
    """Optimizer that computes a moving average of the variables.
    Empirically it has been found that using the moving average of the trained
    parameters of a deep network is better than using its trained parameters
    directly. This optimizer allows you to compute this moving average and swap
    the variables at save time so that any code outside of the training loop
    will use by default the average values instead of the original ones.
    Example of usage:
    ```python
    opt = tf.keras.optimizers.SGD(learning_rate)
    opt = tfa.optimizers.MovingAverage(opt)
    ```
    """

    def __init__(self,
                 optimizer: Union[tf.keras.optimizers.Optimizer, str],
                 sequential_update: bool = True,
                 average_decay = 0.99,
                 num_updates: Optional[str] = None,
                 name: str = "MovingAverage",
                 **kwargs):
        r"""Construct a new MovingAverage optimizer.
        Args:
            optimizer: str or `tf.keras.optimizers.Optimizer` that will be
                used to compute and apply gradients.
            sequential_update: Bool. If False, will compute the moving average
                at the same time as the model is updated, potentially doing
                benign data races. If True, will update the moving average
                after gradient updates.
            average_decay: float. Decay to use to maintain the moving averages
                of trained variables.
            num_updates: Optional count of the number of updates applied to
                variables.
            name: Optional name for the operations created when applying
                gradients. Defaults to "MovingAverage".
            **kwargs: keyword arguments. Allowed to be {`clipnorm`,
                `clipvalue`, `lr`, `decay`}. `clipnorm` is clip gradients by
                norm; `clipvalue` is clip gradients by value, `decay` is
                included for backward compatibility to allow time inverse
                decay of learning rate. `lr` is included for backward
                compatibility, recommended to use `learning_rate` instead.
        """
        super().__init__(optimizer, sequential_update, name, **kwargs)
        self._num_updates = num_updates
        if self._num_updates is not None:
            num_updates = tf.cast(
                self._num_updates, tf.float32, name="num_updates")
            average_decay = tf.minimum(
                average_decay, (1.0 + num_updates) / (10.0 + num_updates))

        self._set_hyper("average_decay", average_decay)

    def average_op(self, var, average_var):
        decay = self._get_hyper('average_decay', tf.dtypes.float32)
        return assign_moving_average(average_var, var, decay, False)

    def get_config(self):
        config = {
            'average_decay': self._serialize_hyperparameter('average_decay'),
            'num_updates': self._num_updates,
        }
        base_config = super().get_config()
        return {**base_config, **config}

    def _create_slots(self, var_list):
        self._optimizer._create_slots(var_list=var_list)  # pylint: disable=protected-access
        for var in var_list:
            self.add_slot(var, 'average', var.read_value())

    def _create_slots_bak(self, var_list):
        self._optimizer._create_slots(var_list=var_list)  # pylint: disable=protected-access

        # for tpu can not call var.read_value in init_op
        for var in var_list:
            self.add_slot(var, 'average')
#        if not context.executing_eagerly():
#            for var in var_list:
#                self.add_slot(var, 'average', var.initial_value())
#        else:
#            for var in var_list:
#                self.add_slot(var, 'average', var.read_value())


class WarmupWrap(tf.keras.optimizers.schedules.LearningRateSchedule):
    def __init__(self, scheduler, warm_up_step, name='WarmupWrap'):
        self.scheduler = scheduler
        self.warm_up_step = warm_up_step
        self.name = name

    def get_config(self):
        return {
            "name": self.name,
            "warm_up_step": self.warm_up_step,
            'scheduler': self.scheduler,
        }

    def get_lr(self, step):
        with ops.name_scope_v2(self.name or "WarmupWrap") as name:
            if callable(self.scheduler):
                initial_learning_rate = ops.convert_to_tensor(self.scheduler.initial_learning_rate,
                                                             name="initial_learning_rate")
            else:
                initial_learning_rate = ops.convert_to_tensor(self.scheduler)
            dtype = initial_learning_rate.dtype
            decay_steps = math_ops.cast(self.warm_up_step, dtype)
            global_step_recomp = math_ops.cast(step, dtype)
            p = global_step_recomp / decay_steps
            return math_ops.multiply(
                initial_learning_rate, p, name=name)

    @tf.function
    def __call__(self, step):
        if step > self.warm_up_step:
            if callable(self.scheduler):
                return self.scheduler(step-self.warm_up_step)
            else:
                return self.scheduler
        else:
            return self.get_lr(step)




import os, re
from setuptools import setup, find_packages

PKG = "mautil"
VERSIONFILE = os.path.join(PKG, "_version.py")
verstr = "unknown"
try:
    verstrline = open(VERSIONFILE, "rt").read()
except EnvironmentError:
    pass # Okay, there is no version file.
else:
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        verstr = mo.group(1)
    else:
        raise Exception("unable to find version in {}".format(VERSIONFILE))


setup(
    name = PKG,
    packages = find_packages(exclude=['examples']),
    version = verstr,
    license='MIT',
    description = 'deep learning training framework based on pytorch',
    author = 'Tom Ma',
    author_email = '40295257@qq.com',
    url = 'https://github.com/world2vec/mautil',
    keywords = ['deep learning framework', 'pytorch'],
    install_requires=[
        'torch>=1.6',
        'tqdm',
        'psutil',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Programming Language :: Python :: 3',
    ],
)



import os, logging
from collections import namedtuple, defaultdict, OrderedDict
from copy import deepcopy
import pprint
import numpy as np

from . import util

logger = logging.getLogger(__name__)

InputFeature = namedtuple('InputFeature', ['name', 'shape', 'dtype', 'sparse'])
InputFeature.__new__.__defaults__ = (None, None, None, False)


class BasicCFG(object):
    def update(self, cfg):
        for k, v in cfg.items():
            attr = getattr(self, k, None)
            if isinstance(attr, dict):
                #attr.update(v)
                util.update_dict(attr, v)
            else:
                setattr(self, k, v)

    def dump(self, fpath):
        dirname = os.path.dirname(fpath)
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        save_dict = self.__dict__
        util.dump_json(save_dict, fpath, ensure_ascii=False)

    def load(self, fpath):
        cfg = util.load_json(fpath)
        self.update(cfg)

    def copy(self):
        return deepcopy(self)

    def __str__(self):
        return pprint.pformat(self.__dict__, indent=4)

class CFG(BasicCFG):
    N_INF = int(1e16)
    def __init__(self):
        self.seed = 9527
        self.show_cfg = False

        ## lr
        self.lr = 1e-3
        self.n_lr_warmup_step = 0
        self.n_lr_decay_step = self.N_INF
        self.lr_decay_rate = 1.0
        self.lr_scheduler = None
        self.lr_scheduler_paras = {}

        ## optimizer
        self.opt = 'adamW'
        self.opt_paras = {}

        ## deepspeed
        self.use_deepspeed = False
        self.deepspeed = {
            "fp16": {
                "enabled": False,
                #"loss_scale": 0,
                #"loss_scale_window": 1000,
                #"hysteresis": 2,
                #"min_loss_scale": 1
            },

            "zero_optimization": {
                "stage": 2,
                "allgather_partitions": True,
                "allgather_bucket_size": 5e8,
                "overlap_comm": True,
                "reduce_scatter": True,
                "reduce_bucket_size": 5e8,
                "contiguous_gradients": True,
                "cpu_offload": True
            },

           "optimizer": {
             "type": "AdamW",
           },
           "zero_allow_untested_optimizer": True,

        }


        # train
        self.patience = 1
        self.es_min_delta = 0.0005  ## early stop
        self.n_es_epoch = 1  ## early stop
        self.batch_size = 32
        self.val_batch_size = self.batch_size
        self.abnormal_loss_thr = 0.1 # minimum percentage train loss improvement
        self.n_init_epoch = 0
        self.n_save_step = self.N_INF
        self.n_save_epoch = self.N_INF
        self.n_epoch_step = self.N_INF
        self.n_train_step = self.N_INF
        self.weight_decay = 0
        self.validation_steps = None
        self.validation_freq = 1
        self.epochs = 2
        self.scoring = False
        self.save = False  # if save model after fit
        self.save_best = False  # if save  best model
        self.only_validate = False
        self.no_validate = False
        self.use_fp16 = False
        self.save_opt = False
        self.xla_share_model = False
        self.debug = False
        self.dropout = 0
        self.data_dir = '../data'
        self.output_dir = '../data'
        self.model_dir = None # use data_dir if None
        self.swa_start_epoch = 0 # stochastic weight average if >0. pt only.
        self.ema = None
        # gpu
        self.device = None
        # save and restore
        self.restored_epoch = None
        self.saved_epoch = None
        self.n_keep_ckpt = 1
        self.saved_ckpts = []
        self.restore = False
        self.restore_epoch= None
        self.restore_step = None
        self.restore_strict = False
        # predicting
        self.verbose = 32
        self.to_list = True
        self.do_concat = False
        self.batch_keys = []  # key to output for net output dict
        # self distill
        self.sd_weight = 0.0
        self.sd_warmup_step = 1000



class LossHist(defaultdict):
    def __init__(self, decimal_digits=8, *args, **kwargs):
        self._fstr = "{}:{:." + str(decimal_digits) + "f}"
        super(LossHist, self).__init__(list, *args, **kwargs)

    def append(self, losses):
        for name, loss in losses.items():
            self[name].append(loss)

    def get_avg(self):
        avg_loss = OrderedDict()
        keys = sorted(self.keys())
        for key in keys:
            avg_loss[key] = np.mean(self[key])
        return avg_loss

    def get_avg_sum(self):
        avg_loss = self.get_avg()
        return np.sum([v for k, v in avg_loss.items()])

    def avg_output(self):
        avg_loss = self.get_avg()
        #tot = np.sum(v for k, v in avg_loss.items())
        #avg_loss['tot'] = tot
        outstr = ', '.join(self._fstr.format(k, v) for k, v in avg_loss.items())
        return outstr


class Model(object):
    cfg = CFG()

    def __init__(self, name, cfg={}):
        self.name = name
        self.update_cfg = cfg
        self.cfg = deepcopy(self.cfg)
        self.cfg.update(self.update_cfg)
        util.update_deepspeed_cfg(self.cfg)
        self.restored_cfg = None

        self._rs = np.random.RandomState(self.cfg.seed)

        self._model = None

#        if self.cfg.save and not self.cfg.only_validate and not self.cfg.no_train and not self.cfg.restore:
#            fpath = self.gen_fname('cfg.json')
#            self.cfg.dump(fpath)

        if self.cfg.show_cfg:
            logging.info('model cfg is: %s', self.cfg)


    def create_model(self):
        return self

    def fit(self, train_dataset=None, val_dataset=None, **kwargs):
        best_loss = np.inf;
        best_epoch = -1;
        step = 0
        for epoch in range(self.cfg.epochs):
            if not self.cfg.only_validate:
                step = self.fit_epoch(train_dataset, epoch, step, **kwargs)
                if (epoch + 1) % (self.cfg.n_save_epoch) == 0:
                    logger.info('%%%% save model to %s for epoch:%s', self.cfg.output_dir, epoch)
                    self.save(self.cfg.output_dir, epoch=epoch)
            if val_dataset is not None:
                val_loss = self.val_epoch(val_dataset, epoch)
                if val_loss < best_loss:
                    best_loss = val_loss
                    best_epoch = epoch
                if self.should_stop(self.cfg.es, self.cfg.n_es_epoch, best_loss, val_loss, best_epoch, epoch):
                    break
                if self.cfg.only_validate:
                    break
            if not self.cfg.only_validate and step % self.cfg.n_train_step == 0:
                logger.info('total train step %s done', self.cfg.n_train_step)
                break
        if self.cfg.save and (epoch + 1) % (self.cfg.n_save_epoch) != 0 and (step % self.cfg.n_save_step != 0):
            logger.info('%%%% save model to %s', self.cfg.output_dir)
            self.save(self.cfg.output_dir)

    def _should_stop(self, best_val_loss, val_loss, best_epoch=-1, current_epoch=-1):
        stop = False
        if ((best_val_loss - val_loss) / val_loss < -self.cfg.es_min_delta) and (
                current_epoch - best_epoch) >= self.cfg.n_es_epoch:
            stop = True
        return stop


    def gen_fname(self, *paras, data_dir=None):
        """

        :param paras: subdirectories and/or file name
        :return:
        """
        name = self.name
        if self.cfg.debug:
            name = 'debug_' + name
        if data_dir is None:
            data_dir = self.cfg.output_dir
        fname = os.path.join(data_dir, name, *paras)
        return fname

    def restore(self, model_dir=None):
        if model_dir is None:
            fpath = None
        else:
            fpath = os.path.join(model_dir, 'cfg.json')
        self.restore_cfg(fpath)

    def restore_cfg(self, fpath=None):
        if fpath is None:
            fpath = self.gen_fname('cfg.json')
        self.cfg.load(fpath)
        self.restored_cfg = self.cfg.copy()
        self.cfg.update(self.update_cfg)
        if self.cfg.show_cfg:
            logging.info('model cfg after restore is: %s', self.cfg)

    def save(self):
        self.save_cfg()

    def save_cfg(self):
        fpath = self.gen_fname('cfg.json')
        dirname = os.path.dirname(fpath)
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        self.cfg.dump(fpath)

    def save_predict(self, pred, suffix=''):
        fpath = self.gen_fname('pred' + suffix + '.dump')
        util.dump(pred, fpath)
        logger.info('saved:%s', fpath)

    def load_predict(self, suffix=''):
        fpath = self.gen_fname('pred' + suffix + '.dump')
        pred = util.load_dump(fpath)
        return pred

    def predict(self, x):
        pred = self._model.predict(x)
        return pred

    def score(self, x, y):
        pass



import os, logging
from collections import Counter
from multiprocessing import Pool
from functools import partial
from . import util

logger = logging.getLogger(__name__)

class Vocabulary(object):
    fake = {'[PAD]': 0, '[SOS]': 1, '[EOS]': 2, '[UNK]': 3, '[SEP]': 4, '[MASK]': 5}
    num = len(fake)
    for i in range(0, 10):
        fake['[unused{}]'.format(i+1)] = num+i

    def __init__(self, name, vocab_path, min_cnt=5, tokenizer=None):
        self.name = name
        self.vocab_path = vocab_path
        self.size = None
        self.vocab = None
        self.tokenizer = tokenizer
        self.min_cnt = min_cnt

        if self.vocab_path is not None and os.path.isfile(self.vocab_path):
            self.restore()

    @property
    def word2id(self):
        return self.vocab['word2id']

    @property
    def id2word(self):
        return self.vocab['id2word']

    @property
    def word2cnt(self):
        return self.vocab['word2cnt']

    def restore(self):
        logger.info('load vocab from saved file: %s', self.vocab_path)
        self.vocab = util.load_json(self.vocab_path)
        self.size = len(self.vocab['word2id'])

    def _build_vocab(self, cnt):
        word2id = self.fake.copy()
        id2word = [k for k, v in sorted(word2id.items(), key=lambda x:x[1])]
        for word in sorted(cnt):
            if cnt[word] >= self.min_cnt:
                word2id[word] = len(word2id)
                id2word.append(word)
        self.size = len(word2id)
        self.vocab = {'word2id': word2id, 'id2word': id2word, 'word2cnt': dict(cnt)}


    @staticmethod
    def count(sents, tokenizer=None):
        cnt = Counter()
        if tokenizer:
            for sent in sents:
                for word in tokenizer(sent):
                    cnt.update([word])
        else:
            for sent in sents:
                for word in sent:
                    cnt.update([word])
        return cnt

    def update_chinese_bert_vocab(self, vocab):
        word2id = dict()
        word2id['[UNK]'] = vocab['[UNK]']
        unk_words = dict()
        for k, v in self.vocab['word2id'].items():
            if k in vocab:
                word2id[k] = vocab[k]
            else:
                word2id[k] = vocab['[UNK]']
                if k not in self.fake:
                    unk_words[k] = self.vocab['word2cnt'][k]
        if '[CLS]' in vocab:
            word2id['[SOS]'] = vocab['[CLS]']
            word2id['[CLS]'] = vocab['[CLS]']
        if '[SEP]' in vocab:
            word2id['[EOS]'] = vocab['[SEP]']
            word2id['[SEP]'] = vocab['[SEP]']
        if '[MASK]' in vocab:
            word2id['[MASK]'] = vocab['[MASK]']
        unk_words = [x[0] for x in sorted(unk_words.items(), key=lambda x: x[1], reverse=True)]
        logging.info('num of unk words: %s, first 20:%s', len(unk_words), unk_words[0:20])
        #for i, word in enumerate(unk_words[0:99]):
        #    word2id[word] = vocab['[unused{}]'.format(i+1)]

        id2word = {v:k for k, v in word2id.items()}
        self.vocab['word2id'] = word2id
        self.vocab['id2word'] = id2word
        self.size = len(vocab)

    def build_vocab(self, sents, save=True, parallel=0):
        if self.vocab is None:
            logger.info('building vocab')
            if parallel >0:
                num = len(sents)
                batch_size = (num//parallel +1)
                batchs = [sents[i:i+batch_size] for i in range(0, len(sents), batch_size)]
                pool = Pool(parallel)
                cnts = pool.map(partial(self.count, tokenizer=self.tokenizer), batchs)
                cnt = sum(cnts, Counter())
            else:
                cnt = self.count(sents, tokenizer=self.tokenizer)
            self._build_vocab(cnt)
            if save:
                util.dump_json(self.vocab, self.vocab_path)
                logger.info('vocab saved to :%s', self.vocab_path)
        logger.info('vocab size:%s', self.size)

    def seq(self, sents, max_len=None, sos=None, eos=None, add_sos=True, add_eos=True, w2i=None, pad=None):
        if w2i is None:
            w2i = self.vocab['word2id']
        if pad is None:
            pad = w2i['[PAD]']
        add_len = 0
        if add_sos:
            add_len += 1
            sos = sos or w2i['[SOS]']
        if add_eos:
            add_len += 1
            eos = eos or w2i['[EOS]']
        seqs = []
        seq_lens = []
        for sent in sents:
            if self.tokenizer:
                tokens = self.tokenizer(sent)
            else:
                tokens = sent
            inds = [w2i[w] if w in w2i else w2i['[UNK]'] for w in tokens]
            if max_len is not None:
                inds = inds[0:max_len-add_len]
            if add_sos:
                inds = [sos] + inds
            if add_eos:
                inds += [eos]
            seq_lens.append(len(inds))
            if max_len is not None:
                inds += [pad] * (max_len - len(inds))
            seqs.append(inds)
        return seqs, seq_lens

    def deseq(self, seqs):
        id2w = self.vocab['id2word']
        sents = []
        for seq in seqs:
            sent = []
            for word_id in seq:
                sent.append(id2w[word_id])
            sents.append(sent)
        return sents



